/**************************************************************************
 *   nano.c                                                               *
 *                                                                        *
 *   Copyright (C) 1999 Chris Allegretta                                  *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 1, or (at your option)  *
 *   any later version.                                                   *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, write to the Free Software          *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>

#ifdef HAVE_TERMIOS_H
#include <termios.h>
#else
#include <termio.h>
#endif

#include "config.h"
#include "proto.h"
#include "nano.h"

/* What we do when we're all set to exit */
RETSIGTYPE finish(int sigage)
{
    blank_bottombars();
    wrefresh(bottomwin);
    endwin();

    exit(sigage);
}

/* Die (gracefully?) */
void die(char *msg, ...)
{
    va_list ap;

    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);

    /* if we can't save we have REAL bad problems,
     * but we might as well TRY.  FIXME: This should probabally base it
     * off of the current filename */
    write_file("nano.save", 0);

    clear();
    refresh();
    resetty();
    endwin();

    fprintf(stderr, msg);
    fprintf(stderr, "\nBuffer written to 'nano.save'\n");

    exit(1);			/* We have a problem: exit w/ errorlevel(1) */
}

/* Thanks BG, many ppl have been asking for this... */
void *nmalloc(size_t howmuch)
{
    void *r;

    /* Panic save? */

    if (!(r = malloc(howmuch)))
	die("nano: malloc: out of memory!");

    return r;
}

/* Initialize global variables - no better way for now */
void global_init(void)
{
    int i;

    center_x = COLS / 2;
    center_y = LINES / 2;
    current_x = 0;
    current_y = 0;
    editwinrows = LINES - 5 + no_help;
    editwineob = editwinrows - 1;
    fileage = NULL;
    cutbuffer = NULL;
    current = NULL;
    totlines = 0;
    placewewant = 0;
    hblank = nmalloc(COLS + 1);

    /* Thanks BG for this bit... */
    for (i = 0; i <= COLS - 1; i++)
	hblank[i] = ' ';
    hblank[i] = 0;

}

/* Make a copy of a node to a pointer (space will be malloc()ed) */
filestruct *copy_node(filestruct * src)
{
    filestruct *dst;

    dst = nmalloc(sizeof(filestruct));
    dst->data = nmalloc(strlen(src->data) + 1);

    dst->next = src->next;
    dst->prev = src->prev;

    strcpy(dst->data, src->data);
    dst->lineno = src->lineno;

    return dst;
}

/* Unlink a node from the rest of the struct */
void unlink_node(filestruct * fileptr)
{
    if (fileptr->prev != NULL)
	fileptr->prev->next = fileptr->next;

    if (fileptr->next != NULL)
	fileptr->next->prev = fileptr->prev;
}

void delete_node(filestruct * fileptr)
{
    if (fileptr->data != NULL)
	free(fileptr->data);
    free(fileptr);
}

/* Okay, now let's duplicate a whole struct! */
filestruct *copy_filestruct(filestruct * src)
{
    filestruct *dst, *tmp, *head, *prev;

    head = copy_node(src);
    dst = head;			/* Else we barf on copying just one line */
    head->prev = NULL;
    tmp = src->next;
    prev = head;

    while (tmp != NULL) {
	dst = copy_node(tmp);
	dst->prev = prev;
	prev->next = dst;

	prev = dst;
	tmp = tmp->next;
    }

    dst->next = NULL;
    return head;
}

/* Free() a single node */
int free_node(filestruct * src)
{
    if (src == NULL)
	return 0;

    if (src->next != NULL)
	free(src->data);
    free(src);
    return 1;
}

int free_filestruct(filestruct * src)
{
    filestruct *fileptr = src;

    if (src == NULL)
	return 0;

    while (fileptr->next != NULL) {
	fileptr = fileptr->next;
	free_node(fileptr->prev);

#ifdef DEBUG
	fprintf(stderr, "free_node(): free'd a node, YAY!\n");
#endif
    }
    free_node(fileptr);
#ifdef DEBUG
    fprintf(stderr, "free_node(): free'd last node.\n");
#endif

    return 1;
}

int renumber_all(void)
{
    filestruct *temp;
    long i = 1;

    totsize = 0;

    for (temp = fileage; temp != NULL; temp = temp->next) {
	temp->lineno = i++;
	temp->bytes = totsize + strlen(temp->data) + 1;
	totsize = temp->bytes;
    }
    if (totsize > 0)
	totsize--;

    return 0;
}

int renumber(filestruct * fileptr)
{
    filestruct *temp;

    if (fileptr == NULL || fileptr->prev == NULL || fileptr == fileage) {
	renumber_all();
	return 0;
    }
    for (temp = fileptr; temp != NULL; temp = temp->next) {
	temp->lineno = temp->prev->lineno + 1;
	temp->bytes = temp->prev->bytes + strlen(temp->data) + 1;
	totsize = temp->bytes;
    }
    if (totsize > 0)
	totsize--;

    return 0;
}

int update_bytes(filestruct * fileptr)
{
    filestruct *temp;

    if (fileptr == NULL) {
	return 0;
    }
    if (fileptr->prev == NULL) {
	if (fileptr->data != NULL)
	    fileptr->bytes = strlen(fileptr->data) + 1;
	else
	    fileptr->bytes = 0;

	totsize = fileptr->bytes;
	fileptr = fileptr->next;
    }
    for (temp = fileptr; temp != NULL; temp = temp->next) {
	temp->bytes = temp->prev->bytes + strlen(temp->data) + 1;
	totsize = temp->bytes;
    }

    if (totsize > 0)
	totsize--;
    return 0;
}

/* Fix the memory allocation for a string */
void align(char **strp)
{
    /* There was a serious bug here:  the new address was never
       stored anywhere... */

    *strp = realloc(*strp, strlen(*strp) + 1);
}

/* Load file into edit buffer - takes data from file struct */
void load_file(void)
{
    current = fileage;
    wmove(edit, current_y, current_x);
}

/* What happens when there is no file to open? aiee! */
void new_file(void)
{
    fileage = nmalloc(sizeof(filestruct));
    fileage->data = nmalloc(1);
    strcpy(fileage->data, "");
    fileage->prev = NULL;
    fileage->next = NULL;
    fileage->lineno = 1;
    filebot = fileage;
    edittop = fileage;
    editbot = fileage;
    current = fileage;
    totlines = 1;
}


int read_byte(char *filename, char *input)
{
    static char buf[BUFSIZ];
    static int index = 0;
    static int size = 0;

    if (index == size) {
	index = 0;
	size = read(file, buf, BUFSIZ);
	if (size == -1) {
	    clear();
	    refresh();
	    resetty();
	    endwin();
	    perror(filename);
	}
	if (!size)
	    return 0;
    }
    *input = buf[index++];
    return 1;
}

filestruct *read_line(char *buf, filestruct * prev, int *line1ins)
{
    filestruct *fileptr;

    fileptr = nmalloc(sizeof(filestruct));
    fileptr->data = nmalloc(strlen(buf) + 2);
    strcpy(fileptr->data, buf);

    if (prev) {
	fileptr->bytes = prev->bytes + strlen(fileptr->data) + 1;
    } else {
	fileptr->bytes = strlen(fileptr->data) + 1;
    }

    if (*line1ins) {
	/* Special case, insert with cursor on 1st line. */
	fileptr->prev = NULL;
	fileptr->next = fileage;
	fileptr->lineno = 1;
	*line1ins = 0;
	fileage = fileptr;
    } else if (fileage == NULL) {
	fileage = fileptr;
	fileage->lineno = 1;
	fileage->next = fileage->prev = NULL;
	fileptr = filebot = fileage;
    } else if (prev) {
	fileptr->prev = prev;
	fileptr->next = NULL;
	fileptr->lineno = prev->lineno + 1;
	prev->next = fileptr;
    } else {
	die("read_line: not on first line and prev is NULL");
    }

    return fileptr;
}


int read_file(char *filename)
{
    long size, lines = 0, linetemp = 0;
    char input[2];		/* buffer */
    char *buf;
    long i = 0, bufx = 128;
    filestruct *fileptr = current, *tmp = NULL;
    int line1ins = 0;

    buf = malloc(bufx);

    if (fileptr != NULL && fileptr->prev != NULL) {
	fileptr = fileptr->prev;
	tmp = fileptr;
    } else if (fileptr != NULL && fileptr->prev == NULL) {
	tmp = fileage;
	current = fileage;
	line1ins = 1;
    }
    input[1] = 0;
    /* Read the entire file into file struct */
    while ((size = read_byte(filename, input)) > 0) {
	linetemp = 0;
	if (input[0] == '\n') {
	    fileptr = read_line(buf, fileptr, &line1ins);
	    lines++;
	    buf[0] = 0;
	    i = 0;
	} else {
	    /* Now we allocate a bigger buffer 128 characters at a time.
	       If we allocate a lot of space for one line, we may indeed 
	       have to use a buffer this big later on, so we don't
	       decrease it at all.  We do free it at the end though. */

	    if (i >= bufx - 1) {
		buf = realloc(buf, bufx + 128);
		bufx += 128;
	    }
	    buf[i] = input[0];
	    buf[i + 1] = 0;
	    i++;
	}
	totsize += size;
    }

    /* Did we not get a newline but still have stuff to do? */
    if (buf[0]) {
	fileptr = read_line(buf, fileptr, &line1ins);
	lines++;
	buf[0] = 0;
    }
    /* Did we even GET a file? */
    if (totsize == 0) {
	new_file();
	statusbar("Read %d lines", lines);
	return 1;
    } else {
	/* do the pico thing where we get a spare blank line */
	fileptr = read_line("", fileptr, &line1ins);

	/* Only update totlines so that we report
	 * the number of lines read correctly */
	totlines++;
    }
    if (current != NULL) {
	fileptr->next = current;
	current->prev = fileptr;
	renumber(current);
	/* FIXME, update the bytes portion for the rest of the file also */
	current_x = 0;
	placewewant = 0;
	edit_update(fileptr);
    } else if (fileptr->next == NULL) {
	filebot = fileptr;
	load_file();
    }
    statusbar("Read %d lines", lines);
    totlines += lines;

    free(buf);
    close(file);

    return 1;
}

/* Open the file (and decide if it exists) */
int open_file(char *filename, int insert, int quiet)
{

    if (!strcmp(filename, "") || stat(filename, &fileinfo) == -1) {
	if (insert) {
	    if (!quiet)
		statusbar("\"%s\" not found", filename);
	    return -1;
	} else {
	    /* We have a new file */
	    statusbar("New File");
	    new_file();
	}
    } else if ((file = open(filename, O_RDONLY)) == -1) {
	if (!quiet)
	    statusbar("%s: %s", strerror(errno), filename);
	return -1;
    } else {			/* File is A-OK */

	if (!quiet)
	    statusbar("Reading File");
	read_file(filename);
    }

    return 1;
}

int do_insertfile(void)
{
    int i;

    i = statusq(writefile_list, WRITEFILE_LIST_LEN, "",
		"File to insert [from ./] ");
    if (i != -1) {

#ifdef DEBUG
	fprintf(stderr, "filename is %s", answer);
#endif

	i = open_file(answer, 1, 0);

	dump_buffer(fileage);
	set_modified();
	return i;
    } else {
	statusbar("Cancelled");
	return 0;
    }
}

void usage(void)
{

    printf("Usage: nano [GNU long option] [option] +LINE <file>\n\n");
    printf("Option		Long option		Meaning\n");
    printf
	(" -V 		--version		Print version information and exit\n");
    printf
	(" -c 		--const			Constantly show cursor position\n");
    printf
	(" -h 		--help			Show this message\n");
    printf
	(" -i 		--autoindent		Automatically indent new lines\n");
    printf
	(" -s [prog] 	--speller=[prog]	Enable alternate speller\n");
    printf
	(" -t 		--tempfile		Auto save on exit, don't prompt\n");
    printf
	(" -w 		--nowrap		Don't wrap long lines\n");
    printf
	(" -x 		--nohelp		Don't show help window\n");
    printf(" -z 		--suspend		Enable suspend\n");
    printf
	(" +LINE					Start at line number LINE\n");

    exit(0);
}

void version(void)
{
    printf(" nano version %s by Chris Allegretta\n", VERSION);
}

void page_down_center(void)
{
    if (editbot->next != NULL && editbot->next != filebot) {
	edit_update(editbot->next);
	center_cursor();
    } else if (editbot != filebot) {
	edit_update(editbot);
	center_cursor();
    } else
	while (current != filebot) {
	    current = current->next;
	    current_y++;
	}

    update_cursor();
}

void page_down(void)
{
    if (current == filebot)
	return;

    if (editbot != filebot) {
	current_y = 0;
	current = editbot;
    } else
	while (current != filebot) {
	    current = current->next;
	    current_y++;
	}

    edit_update_top(current);
    update_cursor();
}

void do_home(void)
{
    current_x = 0;
    placewewant = 0;
}

void do_end(void)
{
    current_x = strlen(current->data);
    placewewant = xplustabs();
}

filestruct *make_new_node(filestruct * prevnode)
{
    filestruct *newnode;

    newnode = nmalloc(sizeof(filestruct));
    newnode->data = NULL;

    newnode->prev = prevnode;
    newnode->next = NULL;

    if (prevnode != NULL)
	newnode->lineno = prevnode->lineno + 1;

    return newnode;
}

void do_mark()
{
    if (!mark_isset) {
	statusbar("Mark Set");
	mark_isset = 1;
	mark_beginbuf = current;
	mark_beginx = current_x;
    } else {
	statusbar("Mark UNset");
	mark_isset = 0;
	mark_beginbuf = NULL;
	mark_beginx = 0;
    }
}

/* What happens when we want to go past the bottom of the buffer */
int do_down(void)
{
    if (current->next != NULL) {
	if (placewewant > 0)
	    current_x = actual_x(current->next, placewewant);

	if (current_x > strlen(current->next->data))
	    current_x = strlen(current->next->data);
    } else
	return 0;

    if (current_y < editwineob && current != editbot)
	current_y++;
    else
	page_down_center();

    update_cursor();

    return 1;
}

int page_up_center(void)
{
    if (edittop != fileage) {
	edit_update(edittop);
	center_cursor();
    } else
	current_y = 0;

    update_cursor();

    return 1;
}

void do_up(void)
{
    if (current->prev != NULL) {
	if (placewewant > 0)
	    current_x = actual_x(current->prev, placewewant);

	if (current_x > strlen(current->prev->data))
	    current_x = strlen(current->prev->data);
    }
    if (current_y > 0)
	current_y--;
    else
	page_up_center();

    update_cursor();
    wrefresh(edit);
}

void do_right(void)
{
    if (current_x < strlen(current->data)) {
	current_x++;
    } else {
	if (do_down())
	    current_x = 0;
    }

    placewewant = xplustabs();
}

void do_left(void)
{
    if (current_x > 0)
	current_x--;
    else if (current != fileage) {
	placewewant = 0;
	current_x = strlen(current->prev->data);
	do_up();
    }
    placewewant = xplustabs();
}

/* The user typed a printable character; add it to the edit buffer */
void do_char(char ch)
{
    /* More dangerousness fun =) */
    current->data = realloc(current->data, strlen(current->data) + 2);
    memmove(&current->data[current_x + 1],
	    &current->data[current_x],
	    strlen(current->data) - current_x + 1);
    current->data[current_x] = ch;
    update_line(current);
    do_right();
    update_cursor();
    update_bytes(current);
    if (!no_wrap)
	check_wrap(current);
    set_modified();
    check_statblank();
    keep_cutbuffer = 0;
    if (totsize == 1 || (current == filebot && strlen(filebot->data) > 0)) {
	/* More magic.  If we're trying to overwrite our 'magic
	   extra line', then make a new one */

	filebot->next = malloc(sizeof(filestruct));
	filebot->next->data = nmalloc(1);
	filebot->next->data[0] = 0;
	filebot->next->prev = filebot;
	filebot->next->next = NULL;
	filebot->next->lineno = filebot->lineno + 1;
	filebot->next->bytes = filebot->bytes + 1;
	filebot = filebot->next;
	/* Only update totlines so that we report
	 * the number of lines read correctly */
	totlines++;
    }
}

/* Someone hits return *gasp!* */
void do_enter(filestruct * inptr)
{
    filestruct *new;
    char *tmp;

    new = make_new_node(inptr);

    tmp = &current->data[current_x];
    new->data = nmalloc(strlen(tmp) + 1);
    strcpy(new->data, tmp);
    *tmp = 0;

    new->next = inptr->next;
    new->prev = inptr;
    inptr->next = new;
    if (new->next != NULL)
	new->next->prev = new;
    else
	filebot = new;

    renumber(current);
    current = new;
    current_x = 0;
    align(&current->data);

    if (current_y == editwinrows - 1)
	edit_update(current);

    totlines++;
    set_modified();

    /* Do auto-indenting, like the neolithic Turbo Pascal editor */
    if (autoindent) {
	tmp = current->prev->data;
	if (tmp)
	    while ((*tmp == ' ') || (*tmp == '\t')) {
		do_char(*tmp);
		tmp++;
	    }
    }
}

void do_next_word(void)
{
    filestruct *fileptr;
    int i;

    if (current == NULL)
	return;

    i = current_x;
    for (fileptr = current; fileptr != NULL; fileptr = fileptr->next) {
	if (fileptr == current) {
	    while (isalnum(fileptr->data[i]) && fileptr->data[i] != 0)
		i++;

	    if (fileptr->data[i] == 0) {
		i = 0;
		continue;
	    }
	}
	while (!isalnum(fileptr->data[i]) && fileptr->data[i] != 0)
	    i++;

	if (fileptr->data[i] != 0)
	    break;

	i = 0;
    }
    if (fileptr == NULL)
	current = filebot;
    else
	current = fileptr;

    current_x = i;
    placewewant = xplustabs();
    if (current->lineno >= editbot->lineno)
	edit_update(current);

}

/* Actually wrap a line, called by check_wrap() */
void do_wrap(filestruct * inptr)
{
    int i, j;
    char *tmp, *foo;

    i = actual_x(inptr, COLS);
    if (inptr->data[i] == ' ')
	while (inptr->data[i] == ' ' && i != 0)
	    i--;

    while (inptr->data[i] != ' ' && i != 0)
	i--;
    i++;

    if (samelinewrap) {
	tmp = &current->data[i];
	foo = nmalloc(strlen(tmp) + strlen(current->next->data) + 1);
	strcpy(foo, tmp);
	strcpy(&foo[strlen(tmp)], current->next->data);
	free(current->next->data);
	current->next->data = foo;
	*tmp = 0;
	if (current_x >= i) {
	    current_x = current_x - i;
	    current = current->next;
	}
	align(&current->next->data);
    } else {
	j = current_x;
	current_x = i;
	do_enter(current);
	if (j > i) {
	    current_x = j - i;
	    samelinewrap = 0;
	} else {
	    current_x = j;
	    current = current->prev;
	    samelinewrap = 1;
	}
    }

}

/* Check to see if we've just caused the line to wrap to a new line */
void check_wrap(filestruct * inptr)
{

#ifdef DEBUG
    fprintf(stderr, "check_wrap called with inptr->data=\"%s\"\n",
	    inptr->data);
#endif

    if ((int) strlenpt(inptr->data) <= COLS - 1)
	return;
    else
	do_wrap(inptr);
}

/* Stuff we do when we abort from programs and want to clean up the
 * screen.  This doesnt do much right now.
 */
void do_early_abort(void)
{
    blank_statusbar_refresh();
}

/* Set up the system variables for a search or replace.  Returns -1 on
   abort, 0 on success, and 1 on rerun calling program */
int search_init(void)
{
    int i;
    char buf[135];

    if (last_search[0]) {
	sprintf(buf, " [%s]", last_search);
    } else {
	buf[0] = '\0';
    }

    if (case_sensitive)
	i = statusq(whereis_list, WHEREIS_LIST_LEN, "",
		    "Case Sensitive Search%s", buf);
    else
	i = statusq(whereis_list, WHEREIS_LIST_LEN, "", "Search%s", buf);

    /* Cancel any search, or just return with no previous search */
    if ((i == -1) || (i < 0 && !last_search[0])) {
	statusbar("Search Cancelled");
	reset_cursor();
	return -1;
    } else if (i == -2) {	/* Same string */
	strncpy(answer, last_search, 132);
    } else if (i == 0) {	/* They entered something new */
	strncpy(last_search, answer, 132);

	/* Blow away last_replace because they entered a new search
	   string....uh, right? =) */
	last_replace[0] = '\0';
    } else if (i == NANO_CASE_KEY) {	/* They want it case sensitive */
	case_sensitive = 1 - case_sensitive;
	return 1;
    } else {			/* First line key, etc. */
	do_early_abort();
	return -3;
    }

    return 0;
}

filestruct *findnextstr(int quiet, filestruct * begin, char *needle)
{
    filestruct *fileptr;
    char *searchstr, *found = NULL, *tmp;

    fileptr = current;

    searchstr = &current->data[current_x + 1];
    /* Look for searchstr until EOF */
    while (fileptr != NULL &&
	   (found = strstrwrapper(searchstr, needle)) == NULL) {
	fileptr = fileptr->next;

	if (fileptr == begin)
	    return NULL;

	if (fileptr != NULL)
	    searchstr = fileptr->data;
    }

    /* If we're not at EOF, we found an instance */
    if (fileptr != NULL) {
	current = fileptr;
	current_x = 0;
	for (tmp = fileptr->data; tmp != found; tmp++)
	    current_x++;

	edit_update(current);
	reset_cursor();
    } else {			/* We're at EOF, go back to the top, once */

	fileptr = fileage;

	while (fileptr != current && fileptr != begin &&
	       (found = strstrwrapper(fileptr->data, needle)) == NULL)
	    fileptr = fileptr->next;

	if (fileptr == begin) {
	    if (!quiet)
		statusbar("\"%s\" not found", needle);

	    return NULL;
	}
	if (fileptr != current) {	/* We found something */
	    current = fileptr;
	    current_x = 0;
	    for (tmp = fileptr->data; tmp != found; tmp++)
		current_x++;

	    edit_update(current);
	    reset_cursor();

	    if (!quiet)
		statusbar("Search Wrapped");
	} else {		/* Nada */

	    if (!quiet)
		statusbar("\"%s\" not found", needle);
	    return NULL;
	}
    }

    return fileptr;
}

/* Search for a string */
void do_search(void)
{
    int i;
    filestruct *fileptr = current;

    if ((i = search_init()) == -1) {
	current = fileptr;
	return;
    } else if (i == -3) {
	return;
    } else if (i == 1) {
	do_search();
	return;
    }
    findnextstr(0, current, answer);
}

void print_replaced(int num)
{
    if (num > 1)
	statusbar("Replaced %d occurences", num);
    else if (num == 1)
	statusbar("Replaced 1 occurence");
}

void replace_abort(void)
{
    keep_cutbuffer = 0;
    bottombars(main_list, MAIN_LIST_LEN);
    reset_cursor();
}

/* Search for a string */
void do_replace(void)
{
    int i, replaceall = 0, numreplaced = 0, beginx;
    filestruct *fileptr, *begin;
    char *tmp, *copy, prevanswer[132] = "";

    if ((i = search_init()) == -1) {
	statusbar("Replace Cancelled");
	replace_abort();
	return;
    } else if (i == 1) {
	do_replace();
	return;
    } else if (i == -3) {
	replace_abort();
	return;
    }
    strncpy(prevanswer, answer, 132);

    if (strcmp(last_replace, "")) {	/* There's a previous replace str */
	i = statusq(replace_list, REPLACE_LIST_LEN, "",
		    "Replace with [%s]", last_replace);

	if (i == -1) {		/* Aborted enter */
	    strncpy(answer, last_replace, 132);
	    statusbar("Replace Cancelled");
	    replace_abort();
	    return;
	} else if (i == 0)	/* They actually entered something */
	    strncpy(last_replace, answer, 132);
	else if (i == NANO_CASE_KEY) {	/* They asked for case sensitivity */
	    case_sensitive = 1 - case_sensitive;
	    do_replace();
	    return;
	} else {		/* First page, last page, for example could get here */

	    do_early_abort();
	    replace_abort();
	    return;
	}
    } else {			/* last_search is empty */

	i = statusq(replace_list, REPLACE_LIST_LEN, "", "Replace with");
	if (i == -1) {
	    statusbar("Replace Cancelled");
	    reset_cursor();
	    replace_abort();
	    return;
	} else if (i == 0)	/* They entered something new */
	    strncpy(last_replace, answer, 132);
	else if (i == NANO_CASE_KEY) {	/* They want it case sensitive */
	    case_sensitive = 1 - case_sensitive;
	    do_replace();
	    return;
	} else {		/* First line key, etc. */

	    do_early_abort();
	    replace_abort();
	    return;
	}
    }

    /* save where we are */
    begin = current;
    beginx = current_x;

    while (1) {

	if (replaceall)
	    fileptr = findnextstr(1, begin, prevanswer);
	else
	    fileptr = findnextstr(0, begin, prevanswer);

	/* No more matches.  Done! */
	if (!fileptr)
	    break;

	/* If we're here, we've found the search string */
	if (!replaceall)
	    i = do_yesno(1, 1, "Replace this instance?");

	if (i > 0 || replaceall) {	/* Yes, replace it!!!! */
	    if (i == 2)
		replaceall = 1;

	    /* Create buffer */
	    copy = nmalloc(strlen(current->data) - strlen(last_search) +
			   strlen(last_replace) + 1);

	    /* Head of Original Line */
	    strncpy(copy, current->data, current_x);
	    copy[current_x] = 0;

	    /* Replacement Text */
	    strcat(copy, last_replace);

	    /* The tail of the original line */
	    /*  This may expose other bugs, because it no longer
	       goes through each character on the string
	       and tests for string goodness.  But because
	       we can assume the invariant that current->data
	       is less than current_x + strlen(last_search) long,   
	       this should be safe.  Or it will expose bugs ;-) */
	    tmp = current->data + current_x + strlen(last_search);
	    strcat(copy, tmp);

	    /* Cleanup */
	    free(current->data);
	    current->data = copy;

	    edit_refresh();
	    set_modified();
	    numreplaced++;
	} else if (i == -1)	/* Abort, else do nothing and continue loop */
	    break;
    }

    current = begin;
    current_x = beginx;
    renumber_all();
    edit_update(current);
    print_replaced(numreplaced);
    replace_abort();
}


int page_up(void)
{
    if (current == fileage)
	return 0;

    current_y = 0;
    edit_update_bot(edittop);
    update_cursor();

    return 1;
}

void delete_buffer(filestruct * inptr)
{
    if (inptr != NULL) {
	delete_buffer(inptr->next);
	free(inptr->data);
	free(inptr);
    }
}

void do_backspace(void)
{
    filestruct *previous;

    if (current != NULL && current_x == 0 && current == filebot
	&& totsize > 1) {
	do_left();
	return;
    }
    if (current_x != 0) {
	/* Let's get dangerous */
	memmove(&current->data[current_x - 1], &current->data[current_x],
		strlen(current->data) - current_x + 2);
#ifdef DEBUG
	fprintf(stderr, "current->data now = \"%s\"\n", current->data);
#endif
	align(&current->data);
	update_bytes(current);
	do_left();
    } else {
	if (current == fileage)
	    return;		/* Can't delete past top of file */

	previous = current->prev;
	current_x = strlen(previous->data);
	previous->data = realloc(previous->data,
				 strlen(previous->data) +
				 strlen(current->data) + 1);
	strcat(previous->data, current->data);
	if (current->next != NULL)
	    current->next->prev = current->prev;
	previous->next = current->next;

	unlink_node(current);
	delete_node(current);
	if (current == edittop)
	    page_up();

	/* Ooops, sanity check */
	if (current == filebot && current->prev != NULL)
	    filebot = filebot->prev;

	current = previous;
	renumber(current);
	previous_line();
	totlines--;


#ifdef DEBUG
	fprintf(stderr, "After, data = \"%s\"\n", current->data);
#endif

    }

    set_modified();
    keep_cutbuffer = 0;
}

void do_delete(void)
{
    if (current_x != strlen(current->data)) {
	/* Let's get dangerous */
	memmove(&current->data[current_x], &current->data[current_x + 1],
		strlen(current->data) - current_x + 1);
#ifdef DEBUG
	fprintf(stderr, "current->data now = \"%s\"\n", current->data);
#endif
	align(&current->data);
	update_bytes(current);

    } else if (current->next != NULL) {
	current_x = 0;
	current = current->next;
	do_backspace();
    } else
	return;

    set_modified();
    keep_cutbuffer = 0;
}

void do_gotoline(long defline)
{
    long line, i = 1, j = 0;
    filestruct *fileptr;

    if (defline > 0)		/* We already know what line we want to go to */
	line = defline;
    else {			/* Ask for it */

	j =
	    statusq(replace_list, REPLACE_LIST_LEN, "",
		    "Enter line number");
	if (j == -1) {
	    statusbar("Aborted");
	    reset_cursor();
	    return;
	} else if (j != 0) {
	    do_early_abort();
	    return;
	}
	if (!strcmp(answer, "$")) {
	    current = filebot;
	    current_x = 0;
	    edit_update(current);
	    reset_cursor();
	    return;
	}
	line = atoi(answer);
    }

    /* Bounds check */
    if (line <= 0) {
	statusbar("Come on, be reasonable");
	return;
    }
    if (line > totlines) {
	statusbar("Only %d lines available, skipping to last line",
		  filebot->lineno);
	current = filebot;
	current_x = 0;
	edit_update(current);
	reset_cursor();
    } else {
	for (fileptr = fileage; fileptr != NULL && i < line; i++)
	    fileptr = fileptr->next;

	current = fileptr;
	current_x = 0;
	edit_update(current);
	reset_cursor();
    }

}

void wrap_reset(void)
{
    samelinewrap = 0;
}

/*
 * Write a file out.  If tmpfile is nonzero, we set the umask to 0600,
 * we don't set the global variable filename to it's name, and don't
 * print out how many lines we wrote on the statusbar.
 */
int write_file(char *name, int tmpfile)
{
    long size, lineswritten = 0;
    char buf[PATH_MAX + 1];
    filestruct *fileptr;
    int mask = 0;
    struct stat st;

    if (!strcmp(name, "")) {
	statusbar("Cancelled");
	return -1;
    }

    titlebar();
    fileptr = fileage;

    memset(buf, 0x00, PATH_MAX + 1);
    if (strlen(name) > (PATH_MAX - 7)) {
	statusbar("Could not open file: Path length exceeded.");
	return -1;
    }

    strcat(buf, name);
    strcat(buf, ".XXXXXX");
    if ((file = mkstemp(buf)) == -1) {
	statusbar("Could not open file for writing: %s", strerror(errno));
	return -1;
    }
    dump_buffer(fileage);
    while (fileptr != NULL && fileptr->next != NULL) {
	/* Cleanup that extra line we added at the beginning */
	if (filebot == fileptr && fileptr->data[0] == '\0')
	    break;

	size = write(file, fileptr->data, strlen(fileptr->data));
	if (size == -1) {
	    statusbar("Could not open file for writing: %s",
		      strerror(errno));
	    return -1;
	} else {
#ifdef DEBUG
	    fprintf(stderr, "Wrote >%s\n", fileptr->data);
#endif
	}
	write(file, "\n", 1);

	fileptr = fileptr->next;
	lineswritten++;
    }

    if (fileptr != NULL) {
	size = write(file, fileptr->data, strlen(fileptr->data));
	if (size == -1) {
	    statusbar("Could not open file for writing: %s",
		      strerror(errno));
	    return -1;
	}
    }

    if (close(file) == -1) {
	statusbar("Could not close %s: %s", name, strerror(errno));
	unlink(buf);
	return -1;
    }

    if (stat(name, &st) == -1) {
	/* Use default umask as file permisions if file is a new file. */
	mask = umask(0);
	umask(mask);

/*	if (tmpfile) *//* We don't want anyone reading our tempfile!  */
/*	    mask = 0600;
	else */
	mask ^= 0666;

    } else {
	/* Use permissions from file we are overwriting. */
	mask = st.st_mode;
	if (unlink(name) == -1) {
	    if (errno != ENOENT) {
		statusbar("Could not open %s for writing: %s",
			  name, strerror(errno));
		unlink(buf);
		return -1;
	    }
	}
    }

    if (link(buf, name) == -1) {
	statusbar("Could not open %s for writing: %s",
		  name, strerror(errno));
	unlink(buf);
	return -1;
    }
    unlink(buf);

    if (chmod(name, mask) == -1) {
	statusbar("Could not set permissions %o on %s: %s",
		  mask, name, strerror(errno));
    }

    if (!tmpfile) {
	strncpy(filename, name, 132);
	statusbar("Wrote %d lines", lineswritten);
    }
    modified = 0;
    titlebar();
    return 1;
}


int do_writeout(int exiting)
{
    int i;

    if ((exiting) && (temp_opt) && (filename)) {
	i = 1;
	strncpy(answer, filename, 132);
    } else {
	i = statusq(writefile_list, WRITEFILE_LIST_LEN, filename,
		    "File Name to write");
    }
    if (i != -1) {

#ifdef DEBUG
	fprintf(stderr, "filename is %s", answer);
#endif

	i = write_file(answer, 0);

	return i;
    } else {
	statusbar("Cancelled");
	return 0;
    }
}

/* Stuff we want to do when we exit the spell program one of its many ways */
void exit_spell(char *tmpfilename, char *foo)
{
    free(foo);

    if (remove(tmpfilename) == -1)
	statusbar("Error deleting tempfile, ack!");
}

/*
 * This is Chris' very ugly spell function.  Someone please make this
 * better =-)
 */
void do_spell(void)
{
    char *temp, *foo;
    int i;

    if ((temp = tempnam(0, "nano.")) == NULL) {
	statusbar("Could not create a temporary filename: %s",
		  strerror(errno));
	return;
    }

    if (write_file(temp, 1) == -1)
	return;

    if (alt_speller) {
	foo = nmalloc(strlen(temp) + strlen(alt_speller) + 2);
	sprintf(foo, "%s %s", alt_speller, temp);
    } else {
	foo = nmalloc(strlen(temp) + 8);
	sprintf(foo, "spell %s", temp);
    }

    endwin();
    resetty();
    if (alt_speller) {
	if ((i = system(foo)) == -1 || i == 32512) {
	    statusbar("Could not invoke spell program \"%s\"",
		      alt_speller);
	    exit_spell(temp, foo);
	    return;
	}
    } else if ((i = system(foo)) == -1 || i == 32512) {	/* Why 32512? I dont know! */
	/* Okay, try ispell for giggles... */
	sprintf(foo, "ispell %s", temp);
	if ((i = system(foo)) == -1 || i == 32512) {
	    statusbar
		("Could not invoke spell program, tried 'spell' and 'ispell'");
	    exit_spell(temp, foo);
	    return;
	}
    }
    initscr();

    free_filestruct(fileage);
    global_init();
    open_file(temp, 0, 1);
    set_modified();
    exit_spell(temp, foo);
    statusbar("Finished checking spelling");

}

void do_exit(void)
{
    int i;

    if (!modified)
	finish(0);

    if (temp_opt) {
	i = 1;
    } else {
	i =
	    do_yesno(0, 0,
		     "Save modified buffer (ANSWERING \"No\" WILL DESTROY CHANGES) ? ");
    }

#ifdef DEBUG
    dump_buffer(fileage);
#endif

    if (i == 1) {
	if (do_writeout(1) > 0)
	    finish(0);
    } else if (i == 0)
	finish(0);
    else
	statusbar("Cancelled");

    bottombars(main_list, MAIN_LIST_LEN);
}

void handle_sigwinch(int s)
{
    char *tty = NULL;
    int fd = 0;
    int result = 0;
    int i = 0;
    struct winsize win;

    tty = ttyname(0);
    if (!tty)
	return;
    fd = open(tty, O_RDWR);
    if (fd == -1)
	return;
    result = ioctl(fd, TIOCGWINSZ, &win);
    if (result == -1)
	return;


    COLS = win.ws_col;
    LINES = win.ws_row;

    center_x = COLS / 2;
    center_y = LINES / 2;
    editwinrows = LINES - 5 + no_help;
    editwineob = editwinrows - 1;

    free(hblank);
    hblank = nmalloc(COLS + 1);

    for (i = 0; i <= COLS - 1; i++)
	hblank[i] = ' ';
    hblank[i] = 0;


    resizeterm(LINES, COLS);
    if (wresize(topwin, 2, COLS) == ERR)
	die("Cannot resize top win");
    if (mvwin(topwin, 0, 0) == ERR)
	die("Cannot move top win");
    if (wresize(edit, editwinrows, COLS) == ERR)
	die("Cannot resize edit win");
    if (mvwin(edit, 2, 0) == ERR)
	die("Cannot move edit win");
    if (wresize(bottomwin, 3 - no_help, COLS) == ERR)
	die("Cannot resize bottom win");
    if (mvwin(bottomwin, LINES - 3 + no_help, 0) == ERR)
	die("Cannot move bottom win");


    editbot = edittop;

    for (i = 0; (i <= editwineob) && (editbot->next != NULL)
	 && (editbot->next != filebot); i++)
	editbot = editbot->next;

    if (current_y > editwineob) {
	edit_update(editbot);
    }
    erase();
    refresh();
    total_refresh();
}

int main(int argc, char *argv[])
{
    int optchr;
    int kbinput;		/* Input from keyboard */
    long startline = 0;		/* Line to try and start at */
    struct sigaction act;	/* For our lovely signals */
    int option_index = 0;
    int tmpkey = 0;
    struct option long_options[] = {
	{"version", 0, 0, 'V'},
	{"const", 0, 0, 'c'},
	{"suspend", 0, 0, 'z'},
	{"nowrap", 0, 0, 'w'},
	{"nohelp", 0, 0, 'x'},
	{"help", 0, 0, 'h'},
	{"autoindent", 0, 0, 'i'},
	{"tempfile", 0, 0, 't'},
	{"speller", 1, 0, 's'},
	{0, 0, 0, 0}
    };

    while ((optchr = getopt_long(argc, argv, "?Vchis:twxz",
				 long_options, &option_index)) != EOF) {
	switch (optchr) {
	case 'V':
	    version();
	    exit(0);
	case 'c':
	    constupdate = 1;
	    break;
	case 'h':
	case '?':
	    usage();
	    exit(0);
	case 'i':
	    autoindent = 1;
	    break;
	case 's':
	    alt_speller = nmalloc(strlen(optarg) + 1);
	    strcpy(alt_speller, optarg);
	    break;
	case 't':
	    temp_opt = 1;
	    break;
	case 'w':
	    no_wrap = 1;
	    break;
	case 'x':
	    no_help = 1;
	    break;
	case 'z':
	    suspend = 1;
	    break;
	default:
	    usage();
	    exit(0);
	}
    }

    filename = nmalloc(PATH_MAX);
    strcpy(filename, "");

    /* See if there's a non-option in argv (first non-option is the
       filename, if +LINE is not given) */
    if (argc == 1 || argc <= optind)
	strcpy(filename, "");
    else {
	/* Look for the +line flag... */
	if (argv[optind][0] == '+') {
	    startline = atoi(&argv[optind][1]);
	    optind++;
	    if (argc == 1 || argc <= optind)
		strcpy(filename, "");
	    else
		strncpy(filename, argv[optind], 132);
	} else
	    strncpy(filename, argv[optind], 132);

    }
    initscr();
    savetty();
    nonl();
    cbreak();
    noecho();
    timeout(0);

    /* Set up some global variables */
    global_init();

    /* Trap SIGINT and SIGQUIT  cuz we want them to do useful things. */
    act.sa_handler = SIG_IGN;
    sigaction(SIGINT, &act, NULL);
    sigaction(SIGQUIT, &act, NULL);

    if (!suspend)
	sigaction(SIGTSTP, &act, NULL);

    act.sa_handler = handle_sigwinch;
    sigaction(SIGWINCH, &act, NULL);

#ifdef DEBUG
    fprintf(stderr, "Main: set up windows\n");
#endif

    /* Setup up the main text window */
    edit = newwin(editwinrows, COLS, 2, 0);
    keypad(edit, TRUE);

    /* And the other windows */
    topwin = newwin(2, COLS, 0, 0);
    bottomwin = newwin(3 - no_help, COLS, LINES - 3 + no_help, 0);
    keypad(bottomwin, TRUE);

#ifdef DEBUG
    fprintf(stderr, "Main: bottom win\n");
#endif
    /* Set up up bottom of window */
    bottombars(main_list, MAIN_LIST_LEN);

#ifdef DEBUG
    fprintf(stderr, "Main: open file\n");
#endif

    titlebar();
    if (argc == 1)
	new_file();
    else
	open_file(filename, 0, 0);

    if (startline > 0)
	do_gotoline(startline);
    else
	edit_update(fileage);

    edit_refresh();
    reset_cursor();

    while (1) {
	raw();

	kbinput = wgetch(edit);
	noraw();
	cbreak();
	if (kbinput == 27) {	/* Grab Alt-key stuff first */
	    switch (kbinput = wgetch(edit)) {
	    case NANO_ALT_GOTO_KEY:
	    case NANO_ALT_GOTO_KEY - 32:
		kbinput = NANO_GOTO_KEY;
		break;
	    case NANO_ALT_REPLACE_KEY:
	    case NANO_ALT_REPLACE_KEY - 32:
		kbinput = NANO_REPLACE_KEY;
		break;
	    case 91:
		switch (kbinput = wgetch(edit)) {
		case 'A':
		    kbinput = KEY_UP;
		    break;
		case 'B':
		    kbinput = KEY_DOWN;
		    break;
		case 'C':
		    kbinput = KEY_RIGHT;
		    break;
		case 'D':
		    kbinput = KEY_LEFT;
		    break;
		case 'H':
		    kbinput = KEY_HOME;
		    break;
		case 'F':
		    kbinput = KEY_END;
		    break;
		case 49:	/* X window F-keys */
		    tmpkey = wgetch(edit);
		    kbinput = KEY_F(tmpkey) - 48;
		    wgetch(edit);	/* Junk character */
		    break;
		case 53:	/* page up */
		    kbinput = KEY_PPAGE;
		    break;
		case 54:	/* page down */
		    kbinput = KEY_NPAGE;
		    break;

		default:
#ifdef DEBUG
		    fprintf(stderr, "I got Alt-[-%c! (%d)\n",
			    kbinput, kbinput);
#endif
		    break;
		}
		break;
	    default:
#ifdef DEBUG
		fprintf(stderr, "I got Alt-%c! (%d)\n", kbinput, kbinput);
#endif
		break;
	    }
	}
	switch (kbinput) {
	case NANO_EXIT_KEY:
	case NANO_EXIT_FKEY:
	    do_exit();
	    break;
	case NANO_WRITEOUT_KEY:
	case NANO_WRITEOUT_FKEY:
	    do_writeout(0);
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case NANO_INSERTFILE_KEY:
	case NANO_INSERTFILE_FKEY:
	    wrap_reset();
	    do_insertfile();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case NANO_GOTO_KEY:
	    wrap_reset();
	    do_gotoline(0);
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case NANO_WHEREIS_KEY:
	case NANO_WHEREIS_FKEY:
	    wrap_reset();
	    do_search();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    wrefresh(bottomwin);
	    break;
	case NANO_CONTROL_A:
	case KEY_HOME:
	case 362:		/* For rxvt, xterm */
	    do_home();
	    break;
	case NANO_CONTROL_E:
	case KEY_END:
	case 385:		/* For rxvt, aterm */
	    do_end();
	    break;
	case NANO_PREVPAGE_KEY:
	case NANO_PREVPAGE_FKEY:
	case KEY_PPAGE:
	    wrap_reset();
	    current_x = 0;
	    page_up();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case NANO_NEXTPAGE_KEY:
	case NANO_NEXTPAGE_FKEY:
	case KEY_NPAGE:
	    wrap_reset();
	    current_x = 0;
	    page_down();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case NANO_CUT_KEY:
	case NANO_CUT_FKEY:
	    do_cut_text(current);
	    break;
	case NANO_UNCUT_KEY:
	case NANO_UNCUT_FKEY:
	    wrap_reset();
	    do_uncut_text(current);
	    keep_cutbuffer = 0;
	    break;
	case NANO_CONTROL_6:	/* ^^ */
	    do_mark();
	    break;
	case NANO_CURSORPOS_KEY:
	case NANO_CURSORPOS_FKEY:
	    do_cursorpos();
	    break;
	case NANO_SPELL_KEY:
	case NANO_SPELL_FKEY:
	    do_spell();
	    break;
	case NANO_HELP_KEY:
	case NANO_HELP_FKEY:
	    statusbar("Help function not yet implemented, nyah!");
	    break;
	case NANO_JUSTIFY_KEY:
	case NANO_JUSTIFY_FKEY:
	    statusbar
		("Justify function not yet implemented, and I feel fine");
	    break;
	case NANO_REPLACE_KEY:
	    do_replace();
	    keep_cutbuffer = 0;
	    bottombars(main_list, MAIN_LIST_LEN);
	    break;
	case KEY_UP:
	case NANO_CONTROL_P:
	    wrap_reset();
	    do_up();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_DOWN:
	case NANO_CONTROL_N:
	    wrap_reset();
	    do_down();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_LEFT:
	case NANO_CONTROL_B:
	    do_left();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_RIGHT:
	case NANO_CONTROL_F:
	    do_right();
	    update_cursor();
	    keep_cutbuffer = 0;
	    check_statblank();
	    break;
	case KEY_BACKSPACE:
	case NANO_CONTROL_H:
	case 127:
	    do_backspace();
	    break;
	case KEY_DC:
	case NANO_CONTROL_D:
	    do_delete();
	    break;
	case NANO_REFRESH_KEY:
	    total_refresh();
	    break;
	case KEY_ENTER:
	case NANO_CONTROL_M:	/* Enter (^M) - FIXME - must be more 
				   things bound to enter key */
	    do_enter(current);
	    break;
	case 331:		/* Stuff that we don't want to do squat */
	    break;
	case NANO_CONTROL_Z:
	case 407:		/* raw suspend signal?? */
	    noraw();
	    cbreak();
	    raise(SIGTSTP);
	    break;
	case 0:		/* Erg */
	    do_next_word();
	    break;
	case -1:
	case 410:		/* Must ignore this, it gets sent when we resize */
	    break;
	default:
#ifdef DEBUG
	    fprintf(stderr, "I got %c (%d)!\n", kbinput, kbinput);
#endif
	    /* We no longer stop unhandled sequences so that people with
	       odd character sets can type... */

	    do_char(kbinput);

	}

	if (constupdate)
	    do_cursorpos();

	edit_refresh();
	reset_cursor();

    }

    getchar();
    finish(0);

}
